@@ -17,13 +17,10 @@ __attribute__((section(".gnu.linkonce.this_module"))) = {
 static const struct modversion_info ____versions[]
 __used
 __attribute__((section("__versions"))) = {
-	{ 0xa5f26ffd, "module_layout" },
-	{ 0x2af88b74, "spl_debug_msg" },
+	{ 0x51198477, "module_layout" },
+	{ 0xbc32eee7, "spl_panic" },
 	{ 0xe52592a, "panic" },
 	{ 0xb4390f9a, "mcount" },
-	{ 0xd1fe48c6, "spl_debug_bug" },
-	{ 0x4958b7b5, "spl_setup" },
-	{ 0x1f4eafab, "spl_cleanup" },
 };
 
 static const char __module_depends[]
@@ -32,4 +29,4 @@ __attribute__((section(".modinfo"))) =
 "depends=spl";
 
 
-MODULE_INFO(srcversion, "5835515F233A845D5B11799");
+MODULE_INFO(srcversion, "C188A8AF7758C6D5506AE88");
