@@ -17,20 +17,18 @@ __attribute__((section(".gnu.linkonce.this_module"))) = {
 static const struct modversion_info ____versions[]
 __used
 __attribute__((section("__versions"))) = {
-	{ 0xa5f26ffd, "module_layout" },
+	{ 0x51198477, "module_layout" },
 	{ 0x25ec1b28, "strlen" },
 	{ 0xe2d5255a, "strcmp" },
 	{ 0xb4390f9a, "mcount" },
 	{ 0x85abc85f, "strncmp" },
 	{ 0xf0fdf6cb, "__stack_chk_fail" },
-	{ 0x4958b7b5, "spl_setup" },
-	{ 0x1f4eafab, "spl_cleanup" },
 };
 
 static const char __module_depends[]
 __used
 __attribute__((section(".modinfo"))) =
-"depends=spl";
+"depends=";
 
 
-MODULE_INFO(srcversion, "943EB776D6CC4E1BCED8C36");
+MODULE_INFO(srcversion, "5DC8CE3A830A36341CDAF7F");
